/*
 * Created on November 9, 2005, 3:37 PM
 *
 */

package com.sunlabs.archsim.fleet.ships.adder;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;

/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/**
 *
 * @author mh193489
 */
public class AdderAction extends Action {
        private Delay adderDelay= AdderDefaults.ADDER_DELAY_DEF;
                
        Function out= new AddData();
        Function ackIn= new AckFunction();

        InputTerminal in1;
        InputTerminal in2;
        
    	public AdderAction(String name, CompositeEntity parent) {
		super(name, parent);
		outputFunctions.put(AdderDefaults.OUT, out);
 		outputFunctions.put(AdderDefaults.ACK_IN1, ackIn);
 		outputFunctions.put(AdderDefaults.ACK_IN2, ackIn);
                
	} // AdderAction
	
	public void setDelay(Delay d) {
		adderDelay= (Delay)d.clone();
		out.setDelay(adderDelay);
	}

        public boolean selfCheck() {
		boolean OK= true;
		boolean check;

                check= (this.getTerminal(AdderDefaults.IN1) != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal in1");
		} else {
                    in1= (InputTerminal)this.getTerminal(AdderDefaults.IN1);
                }

                check= (this.getTerminal(AdderDefaults.IN2) != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal in2");
		} else {
                    in2= (InputTerminal)this.getTerminal(AdderDefaults.IN2);
                }

                check= (this.getTerminal(AdderDefaults.ACK_OUT) != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal ackOut");
		}
                
		// also run the self check on the superclass
		check= super.selfCheck();
		OK= OK && check;
		
		return OK;
        }
        
        private class AddData extends Function {
		
		public AddData() {
			super(AdderDefaults.ADD_DATA);
			functionDelay= adderDelay;
		}
		
		public Object value() {
                        // get data from input terminals
                        Data data1= ((Payload)in1.getData()).data;
                        Data data2= ((Payload)in2.getData()).data;
                        
                        // check if input data is valid
                        if(data1.valid && data2.valid) {
                            // both inputs valid, hence compute the sum
                            Integer sum= (Integer)data1.value + (Integer)data2.value;
                            //check for over or underflow
                            if (sum <= AdderDefaults.UNDERFLOW) {
                                return new Data(AdderDefaults.MINUS_INFINITY,false);
                                
                            } else if (AdderDefaults.OVERFLOW <= sum) {
                                return new Data(AdderDefaults.PLUS_INFINITY, false);
                                
                            } else {
                                return new Data(sum, true);
                            }
                            
                        } else {
                            // at least one input invalid, hence return invalid sum
                            return new Data(AdderDefaults.OOB, false);
                        }
		} // value
	} // AddData

}


