/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Jan 21, 2005
 */
package com.sunlabs.archsim.gasp.terminals;

/**
 * @author ib27688
 *
 * This class represents an input terminal that connects to the
 * state that provide the data for the n-th argument when an
 * action fires and sets the data on its output side. The number
 * of the argument is provided at the creation of the terminal.
 * 
 * The default role is SELF_RESETTING_INPUT.
 */

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;

public class NumberedInput extends InputTerminal {

	/** the number of the argument that is represented by this input */
	protected int myNumber;
	
	/**
	 * Constructor.
	 * @param name Terminal name.
	 * @param n the numnber of the argument to the function computed
	 * 			when an attached action fires.
	 */
	public NumberedInput(String name, int n) {
		super(name);
		myNumber= n;
	} // NumberedInput constructor

	public NumberedInput(String name, int n, CompositeEntity g) {
		this(name, n);
		setParent(g);
	} // NumberedInput constructor

	
	/** Accessor method for the input number */
	public int number() {
		return myNumber;
	} // number
		
} // NumberedInput
